unit
// OPALESCENCE
// Version: 20091024
//
// A basic GUI toolkit for Cobra
//
// Copyright  2008, 2009 Rob Haskell <flyingcucco at googlemail dot com>
//
// Permission to use, copy, modify, and/or distribute this software for any
// purpose with or without fee is hereby granted, provided that the above
// copyright notice and this permission notice appear in all copies.
//
// THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
// WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
// ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
// ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
// OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

uses
    //works with cobra3d, cobra2d, pure2d
    pure2d

const    
    //widget modes          //| BOOL  | VALUE | TEXT  | LABEL | ICON  | GROUP | HEIGHT| ( R=Read only  )
    WIDGET_LABEL    = 0     //|       | W     |       | X     | W     |       | W     | ( W=Write only )
    WIDGET_CHECK    = 1     //| X     |       |       | X     | W     |       | W     | ( X=Read/Write )
    WIDGET_RADIO    = 2     //| X     |       |       | X     | W     | W     | W     |
    WIDGET_BUTTON   = 3     //|       |       |       | X     | W     |       | W     | Other properties
    WIDGET_TOGGLE   = 4     //| X     |       |       | X     | W     | W     | W     | available for all
    WIDGET_TEXT     = 5     //|       | X     | X     |       |       |       | W     | widget modes.      
    WIDGET_SLIDE    = 6     //|       | X     |       |       |       |       |       |
    WIDGET_DROP     = 7     //|       | X     | R     |       |       |       |       |
    WIDGET_LIST     = 8     //|       | X     | R     |       |       |       | W     |
    
    //widget states   Check/Radio   Slide/Drop
    WIDGET_CLEAR    = 0             //0
    WIDGET_SET      = 1     
    WIDGET_DISABLED = 2             //1
    WIDGET_HOVER    = 4             //2
    WIDGET_HOLD     = 6             //3
    
    //mouse states
    MOUSE_UP        = 0
    MOUSE_CLICK     = 1
    MOUSE_RELEASE   = 2
    MOUSE_DOWN      = MOUSE_CLICK + MOUSE_RELEASE 
    
    //magic numbers
    DEFAULT_Z_ORDER =  2130706432   //$7F000000
    DEFAULT_COLOR   =    16711935   //$00FF00FF, means undefined
    DEFAULT_NUMBER  = -2147483648   //$80000000, means undefined
    DEFAULT_GROUP   =          -1   //means not grouped
    DEFAULT_MIN     =           0
    DEFAULT_MAX     =         100

type gui_listopt = record
    ref             : integer                   //drop list option
    txt             : string
endtype

type gui_widget = record
    ref             : integer                   //unique (to window) ref id - same as list index
    mode            : integer                   //type of widget (see constants)     
    group           : integer = DEFAULT_GROUP   //grouping for radio/push buttons

    value           : integer = DEFAULT_MIN     //numerical data
    value_min       : integer = DEFAULT_MIN     //Min slider value
    value_max       : integer = DEFAULT_MAX     //Max slider value / Max Length of Text / count of listopts
    txt             : string                    //Text data / label / first listopt      
    listopts        : list of gui_listopt

    backcolor       : integer = DEFAULT_COLOR   //background colour, overrides global theme
    textcolor       : integer = DEFAULT_COLOR   //text colour, overrides global theme  
    icon            : element                   //image label shown on widget
    x               : integer                   //position in window
    y               : integer    
    width           : integer                   //size of widget
    height          : integer     
    enabled         : boolean = TRUE            //FALSE greys out widget and disables events
    visible         : boolean = TRUE            //FALSE hides widget
endtype

type gui_window = record
    ref             : integer                   //unique ref id - same as list index
    sprite          : element                   //window graphic
    widgets         : list of gui_widget        //widgets contained in window
    widget_count    : integer                   //number of widgets in window
    
    texture         : element                   //tiled background image
    backcolor       : integer = DEFAULT_COLOR   //background colour, overrides global theme
    x               : integer                   //position on screen
    y               : integer               
    width           : integer                   //size of window
    height          : integer               
    visible         : boolean = TRUE            //FALSE hides window
endtype

var
    //graphics
    gfx_check           : array [8] of element          //check box states
    gfx_radio           : array [8] of element          //radio button states
    gfx_slide           : array [3] of element          //slider states
    gfx_drop            : array [3] of element          //drop box states
    
    //default colours
    opal_window         : integer = $ffc0e0ff   ;export //colour Palette for windows 
    opal_windowborder   : integer = $ff4080c0   ;export 
    opal_label          : integer = $ff000000   ;export  
    opal_labeldisabled  : integer = $ff707070   ;export 
    
    opal_widget         : integer = $ffffffff   ;export //colour Palette for Text boxes etc
    opal_widgetdisabled : integer = $ffc0c0c0   ;export 
    opal_widgetborder   : integer = $ff4080c0   ;export 
    opal_widgetbordis   : integer = $ff707070   ;export 
    opal_widgettext     : integer = $ff000000   ;export 
    opal_widgettextdis  : integer = $ff707070   ;export 
    opal_widgethover    : integer = $ffffc000   ;export
    opal_widgetselect   : integer = $ffffc000   ;export
    
    opal_button         : integer = $ff80c0ff   ;export //colour Palette for buttons etc
    opal_buttondisabled : integer = $ffa0a0a0   ;export 
    opal_buttonborder   : integer = $ff4080c0   ;export 
    opal_buttonbordis   : integer = $ff707070   ;export 
    opal_buttontext     : integer = $ff000000   ;export 
    opal_buttontextdis  : integer = $ff707070   ;export  
    opal_buttonhover    : integer = $ffffc000   ;export 
    opal_buttonhold     : integer = $ff306080   ;export 
    
    //font
    opal_fontface       : string  = "Arial"     ;export         
    opal_fontsize       : integer = 10          ;export
    opal_fontstyle      : integer = 0           ;export
    
    //layout 
    window_zorder       : integer                       //z order of the window sprites
    label_space         : integer                       //width of checkbox/radio graphic    
    text_height         : integer                       //height of a Line of text  
    slide_height        : integer                       //size of slider handle graphic 
    slide_width         : integer                       
    drop_height         : integer                       //size of dropbox button graphic 
    drop_width          : integer                       

    //gui elements
    windows             : list of gui_window       
    window_count        : integer    
    
    drop_box            : element
    drop_box_x          : integer
    drop_box_y          : integer
    drop_box_width      : integer
    drop_box_height     : integer

    //event
    mouse_window        : integer = -1                  //mouse hover target
    mouse_widget        : integer = -1        
    focus_window        : integer = -1                  //selected Text box
    focus_widget        : integer = -1
    old_mouse_down      : integer                       //previous mouse state to determine click/hold/release

procedure OpalInitialise(res_path:string, z_order:integer = DEFAULT_Z_ORDER)
export

//  * Load the graphics.
    
begin                        

    //checkbox graphics 
    gfx_check[WIDGET_CLEAR]                 = LoadImage(res_path+"check.png")
    gfx_check[WIDGET_SET]                   = LoadImage(res_path+"check_set.png")
    gfx_check[WIDGET_DISABLED]              = LoadImage(res_path+"check_disabled.png")
    gfx_check[WIDGET_DISABLED+WIDGET_SET]   = LoadImage(res_path+"check_set_disabled.png")    
    gfx_check[WIDGET_HOVER]                 = LoadImage(res_path+"check_hover.png")
    gfx_check[WIDGET_HOVER+WIDGET_SET]      = LoadImage(res_path+"check_set_hover.png")
    gfx_check[WIDGET_HOLD]                  = LoadImage(res_path+"check_hold.png")
    gfx_check[WIDGET_HOLD+WIDGET_SET]       = LoadImage(res_path+"check_set_hold.png")
    
    //radio button graphics
    gfx_radio[WIDGET_CLEAR]                 = LoadImage(res_path+"radio.png")
    gfx_radio[WIDGET_SET]                   = LoadImage(res_path+"radio_set.png")
    gfx_radio[WIDGET_DISABLED]              = LoadImage(res_path+"radio_disabled.png")
    gfx_radio[WIDGET_DISABLED+WIDGET_SET]   = LoadImage(res_path+"radio_set_disabled.png")
    gfx_radio[WIDGET_HOVER]                 = LoadImage(res_path+"radio_hover.png")
    gfx_radio[WIDGET_HOVER+WIDGET_SET]      = LoadImage(res_path+"radio_set_hover.png")
    gfx_radio[WIDGET_HOLD]                  = LoadImage(res_path+"radio_hold.png")
    gfx_radio[WIDGET_HOLD+WIDGET_SET]       = LoadImage(res_path+"radio_set_hold.png")  

    //leave space in label on checkboxes/radios    
    label_space                             = ImageWidth(gfx_check[0])

    //slider graphics
    //cannot be 'set' so we can just halve the state
    gfx_slide[WIDGET_CLEAR]                 = LoadImage(res_path+"slide.png")
    gfx_slide[WIDGET_DISABLED/2]            = LoadImage(res_path+"slide_disabled.png")
    gfx_slide[WIDGET_HOVER/2]               = LoadImage(res_path+"slide_hover.png")
    gfx_slide[WIDGET_HOLD/2]                = LoadImage(res_path+"slide_hold.png")
    
    //slider layout
    slide_width                             = ImageWidth(gfx_slide[0])
    slide_height                            = ImageHeight(gfx_slide[0])

    //drop graphics
    //cannot be 'set' so we can just halve the state
    gfx_drop[WIDGET_CLEAR]                  = LoadImage(res_path+"drop.png")
    gfx_drop[WIDGET_DISABLED/2]             = LoadImage(res_path+"drop_disabled.png")
    gfx_drop[WIDGET_HOVER/2]                = LoadImage(res_path+"drop_hover.png")
    gfx_drop[WIDGET_HOLD/2]                 = LoadImage(res_path+"drop_hold.png")

    //drop layout
    drop_width                              = ImageWidth(gfx_drop[0])
    drop_height                             = ImageHeight(gfx_drop[0])
    
    //calc font height
    text_height                             = opal_fontsize * 1.5     
    
    //z order - make windows appear above / below other sprites 
    window_zorder                           = z_order
end

function OpalCreateWindow(x:integer, y:integer, width:integer, height:integer, visible:boolean = TRUE, texture:element = NULL, backcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create new GUI window.
//  * Returns unique ref handle.

var
    window : ^gui_window    
    widget : ^gui_widget     
begin
    //create new window
    window           = newitem(windows)
    window.ref       = window_count
    window.x         = x
    window.y         = y
    window.width     = width
    window.height    = height    
    window.sprite    = CreateSprite(width, height)
    window.visible   = visible
    if texture <> NULL then window.texture   = texture  
    window.backcolor = backcolor    
     
    //setup window sprite
    SpriteVisible(window.sprite, visible)  
    PositionSprite(window.sprite, x, y)
    SpriteIndex(window.sprite, window_zorder + window_count)
    SetFont(opal_fontface, opal_fontsize, opal_fontstyle, window.sprite)      
    
    //render window
    DrawWindow(window.ref)
    
    //return ref
    result = window_count       
    inc(window_count)
end

function OpalCreateLabel(window_ref:integer, x:integer, y:integer, width:integer, height:integer, label:string, icon_image:element = NULL, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create a checkbox widget in the window referenced.

begin
    result = CreateWidget(window_ref, WIDGET_LABEL, x, y, width, height, icon_image, 0, 0, 0, label, -1, enabled, visible, textcolor) 
end

function OpalCreateCheck(window_ref:integer, x:integer, y:integer, width:integer, height:integer, label:string, icon_image:element = NULL, value:boolean = FALSE, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create a checkbox widget in the window referenced.

begin
    result = CreateWidget(window_ref, WIDGET_CHECK, x, y, width, height, icon_image, toint(value), -1, 0, label, -1, enabled, visible, textcolor) 
end

function OpalCreateRadio(window_ref:integer, x:integer, y:integer, width:integer, height:integer, label:string, icon_image:element = NULL, value:boolean = FALSE, group:integer = DEFAULT_GROUP, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create a radio button widget.

begin
    result = CreateWidget(window_ref, WIDGET_RADIO, x, y, width, height, icon_image, toint(value), -1, 0, label, group, enabled, visible, textcolor) 
end

function OpalCreateButton(window_ref:integer, x:integer, y:integer, width:integer, height:integer, label:string, icon_image:element = NULL, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create a button widget in the window referenced.

begin
    result = CreateWidget(window_ref, WIDGET_BUTTON, x, y, width, height, icon_image, 0, 0, 0, label, -1, enabled, visible, textcolor) 
end

function OpalCreateToggle(window_ref:integer, x:integer, y:integer, width:integer, height:integer, label:string, icon_image:element = NULL, value:boolean = FALSE, group:integer = DEFAULT_GROUP, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * Create a push button widget in the window referenced.

begin
    result = CreateWidget(window_ref, WIDGET_TOGGLE, x, y, width, height,  icon_image, toint(value), -1, 0, label, group, enabled, visible, textcolor) 
end

function OpalCreateText(window_ref:integer, x:integer, y:integer, width:integer, height:integer, str:string, max_len:integer, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * create Text box widget

begin
    result = CreateWidget(window_ref, WIDGET_TEXT, x, y, width, height, NULL, 0, 0, max_len, str, -1, enabled, visible, textcolor)
end

function OpalCreateSlider(window_ref:integer, x:integer, y:integer, width:integer, value:integer = DEFAULT_MIN, value_min:integer = DEFAULT_MIN, value_max:integer = DEFAULT_MAX, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * create slider widget
//  * slider is fixed height, same height as slider graphic.

begin
    result = CreateWidget(window_ref, WIDGET_SLIDE, x, y, width, slide_height, NULL, value, value_min, value_max, "", -1, enabled, visible, textcolor)
end

function OpalCreateDrop(window_ref:integer, x:integer, y:integer, width:integer, txt:string, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * create dropdown box widget
//  * txt is the first option fom the list
//  * others can be added with OpalAddList

begin
    result = CreateWidget(window_ref, WIDGET_DROP, x, y, width, drop_height, NULL, 0, 0, 1, txt, -1, enabled, visible, textcolor)
end

function OpalCreateList(window_ref:integer, x:integer, y:integer, width:integer, height:integer, txt:string, enabled:boolean = TRUE, visible:boolean = TRUE, textcolor:integer = DEFAULT_COLOR) : integer
export

//  * create list box widget
//  * txt is the first option fom the list
//  * others can be added with OpalAddList

begin
    result = CreateWidget(window_ref, WIDGET_LIST, x, y, width, height, NULL, 0, 0, 1, txt, -1, enabled, visible, textcolor)
end

function OpalReadValue(window_ref:integer, widget_ref:integer) : integer
export

//  * returns the value in the widget. 

var
    window : ^gui_window    
    widget : ^gui_widget 
begin

    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)    
    
    //set result
    result = widget.value  //read value
end

Procedure OpalSetValue(window_ref:integer, widget_ref:integer, value:integer, value_min:integer = DEFAULT_NUMBER, value_max:integer = DEFAULT_NUMBER)
export

//  * Sets the value of the widget referenced to the value specified.

var
    window  : ^gui_window    
    widget  : ^gui_widget
    listopt : ^gui_listopt 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)  
    
    //set value 
    case widget.mode of
    
    (WIDGET_CHECK),(WIDGET_RADIO),(WIDGET_TOGGLE):
        widget.value = value  
        
    (WIDGET_SLIDE):
        widget.value = value
        if value_min > DEFAULT_NUMBER then
            widget.value_min = value_min
            widget.value = Max(value_min, value) //bounding to range
        endif
        if value_max > DEFAULT_NUMBER then
            widget.value_max = value_max 
            widget.value = Min(value_max, value) //bounding to range
        endif
    
    (WIDGET_LABEL):
        widget.value = value
        widget.txt = tostring(value)        

    (WIDGET_TEXT):
        widget.value = value
        widget.txt = tostring(value)
        if value_max > DEFAULT_NUMBER then widget.value_max = value_max
    
    (WIDGET_DROP),(WIDGET_LIST):
        widget.value = value
        widget.value = Max(0, Min(value_max-1, value)) //bounding to number of options
        listopt = indexeditem(widget.listopts, widget.value)
        widget.txt = listopt.txt
    endcase  
    
    //redraw
    DrawWidget(window.ref, widget.ref)                          
end

function OpalReadText(window_ref:integer, widget_ref:integer) : string
export

//  * returns the Text property (str) of the textbox widget. 

var
    window : ^gui_window    
    widget : ^gui_widget 
begin

    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)   
    
    //set result
    result = widget.txt  //read str
end

Procedure OpalSetText(window_ref:integer, widget_ref:integer, txt:string)
export

//  * Sets the Text property of textbox widgets.
//  * Sets the label for widgets that have no string value.

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref) 
     
    //set value
    case widget.mode of
    
    (WIDGET_LABEL),(WIDGET_CHECK),(WIDGET_RADIO),(WIDGET_BUTTON),(WIDGET_TOGGLE):
        widget.txt = txt
        
    (WIDGET_TEXT):
        widget.txt = txt
        widget.value =  toint(txt)
    
    endcase
    
    //redraw
    DrawWidget(window.ref, widget.ref)                          
end

Procedure OpalSetIcon(window_ref:integer, widget_ref:integer, icon_image:element = NULL)
export

//  * Change the image label (icon) on a widget

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)

    //icon  
    case widget.mode of
    
    (WIDGET_LABEL),(WIDGET_CHECK),(WIDGET_RADIO),(WIDGET_BUTTON),(WIDGET_TOGGLE):    
        if icon_image = NULL then
            widget.icon = NULL
        else
            widget.icon = icon_image 
        endif
    
    endcase 
    
    //redraw          
    DrawWidget(window.ref,widget.ref)
end

Procedure OpalSetGroup(window_ref:integer, widget_ref:integer, group:integer = DEFAULT_GROUP)
export

//  * Change wiget group -1 = no group

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)      
    
    //group
    case widget.mode of
    
    (WIDGET_RADIO),(WIDGET_TOGGLE):
        widget.group = group
        
    endcase
end

function OpalAddList(window_ref:integer, widget_ref:integer, txt:string) : integer
export

//  * Add an option to the list for dropboxes/listboxes
//  * Returns index of option 

var
    window      : ^gui_window
    widget      : ^gui_widget
    listopt     : ^gui_listopt
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)

    //fail result
    result = -1 
    
    case widget.mode of
    
    (WIDGET_DROP),(WIDGET_LIST):
        //add item
        listopt = newitem(widget.listopts)
        listopt.txt = txt
        listopt.ref = widget.value_max
        result = listopt.ref            
        inc(widget.value_max)        
    
    endcase  
    
    DrawWidget(window.ref, widget.ref)
end

function OpalClearList(window_ref:integer, widget_ref:integer, txt:string) : integer
export

//  * Clear all the options in a listbox/dropbox and add one new one.
//  * Returns index of option (ie 0) 

var
    window      : ^gui_window
    widget      : ^gui_widget
    listopt     : ^gui_listopt
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)

    //fail result
    result = -1
    
    case widget.mode of
    
    (WIDGET_DROP),(WIDGET_LIST):
        //wipe option list
        loop listopt through widget.listopts
            free(listopt)
        endloop
        //create new option
        listopt = newitem(widget.listopts)
        listopt.txt = txt
        listopt.ref = 0
        result = listopt.ref
        //reset count           
        widget.value_max = 1 
                   
    endcase  
    
    DrawWidget(window.ref, widget.ref)
end

Procedure OpalPositionWindow(window_ref:integer, x:integer, y:integer)
export

//  * move a window around the screen

var
    window : ^gui_window    
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    //position
    window.x = x
    window.y = y 
    PositionSprite(window.sprite, x, y)
end

Procedure OpalPositionWidget(window_ref:integer, widget_ref:integer, x:integer, y:integer)
export

//  * move a widget around in a window

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)
    //position  
    widget.x = x
    widget.y = y
    //redraw whole window             
    DrawWindow(window.ref)
end

Procedure OpalResizeWindow(window_ref:integer, width:integer, height:integer)
export

//  * Change the size of a window.
//  * This is quite inefficient!

var
    window : ^gui_window    
begin
    //set pointer
    window = indexeditem(windows, window_ref)    

    //remove old sprite
    FreeSprite(window.sprite)
    
    //resize
    window.width = width
    window.height = height 
    
    //create new sprite
    window.sprite   = CreateSprite(width, height)
    
    //set other junk
    SpriteVisible(window.sprite, window.visible)  
    PositionSprite(window.sprite, window.x, window.y)
    SpriteIndex(window.sprite, window_zorder + window_count)
    SetFont(opal_fontface, opal_fontsize, opal_fontstyle, window.sprite)
        
    //draw window             
    DrawWindow(window.ref)
end

Procedure OpalResizeWidget(window_ref:integer, widget_ref:integer, width:integer, height:integer)
export

//  * Change the size of a widget.
//  * The label and graphics must fit inside the width and
//  * height, or they will be clipped.

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)  
    
    //position  
    widget.width = width
    widget.height = height    

    case widget.mode of
    
    (WIDGET_SLIDE):
        widget.height = slide_height
        
    (WIDGET_DROP):
        widget.height = drop_height
        
    endcase 
    
    //redraw whole window             
    DrawWindow(window.ref)
end

Procedure OpalEnableWidget(window_ref:integer, widget_ref:integer, enabled:boolean = TRUE)
export

//  * enable or disable widget

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)  
    //set enabled  and redraw
    widget.enabled = enabled
    DrawWidget(window.ref, widget.ref)
end

Procedure OpalShowWindow(window_ref:integer, visible:boolean = TRUE)
export

//  * show or hide window

var
    window : ^gui_window     
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    //set state and sprite
    window.visible = visible
    SpriteVisible(window.sprite, visible)
end

Procedure OpalZOrderWindows (z_order:integer = DEFAULT_NUMBER)
export

//  * apply sprite z-ordering to keep windows on top

var
    window : ^gui_window
    index  : integer
begin
    SpriteAutoSort(FALSE)
    //use global z-order by default
    if z_order = DEFAULT_NUMBER then z_order = window_zorder
    //apply to all windows
    loop window through windows
        SpriteIndex(window.sprite, z_order + index)
        inc(index)
    endloop
    SpriteAutoSort(TRUE)    
end

Procedure OpalShowWidget(window_ref:integer, widget_ref:integer, visible:boolean = TRUE)
export

//  * show or hide widget

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)  
    //set state and redraw
    widget.visible = visible
    DrawWidget(window.ref, widget.ref)
end                                                     

Procedure OpalTextureWindow(window_ref:integer, image:element = NULL)
export

//  * set a background image for a window

var
    window : ^gui_window
begin
    window = indexeditem(windows, window_ref)
    //set texture and redraw 
    window.texture = image    
    DrawWindow(window_ref)    
end

Procedure OpalColorWidget(window_ref:integer, widget_ref:integer, color:integer = DEFAULT_COLOR)
export

//  * set a custom Text color for this widget, default restores theme color

var
    window : ^gui_window    
    widget : ^gui_widget 
begin
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref) 
    //set color
    widget.textcolor = color
    DrawWidget(window.ref, widget.ref)
end 

function OpalEvent(byref window_ref:integer, byref widget_ref:integer) : boolean
export

//  * Event handler processes mouse and keyboard input on the GUI. 
//  * Should be called in program main loop.
//  * Returns TRUE if event occured, FALSE otherwise.
//  * NB mouse events are triggered on mouse release.
//  * If mouse is on widget, its refs are returned in params.
//  * If mouse not on widget/window, refs set to -1.

var
    window      : ^gui_window
    widget      : ^gui_widget
    listopt     : ^gui_listopt
    temp_window : integer
    temp_widget : integer      
    mouse_state : integer 
      
begin

    //mouse state
    if MouseDown(0) then
        mouse_state = MOUSE_CLICK + old_mouse_down
        old_mouse_down = MOUSE_RELEASE
    else
        mouse_state = old_mouse_down
        old_mouse_down = MOUSE_UP 
    endif 

    //*** focus handler *** 
    
    if focus_window > -1
        if drop_box<>NULL
        
            //dropbox 
            if PointInRect(MouseX, MouseY, drop_box_x, drop_box_y, drop_box_width, drop_box_height) then
                if mouse_state = MOUSE_RELEASE then
                    //dropbox select event 
                    
                    //set pointers
                    window = indexeditem(windows, focus_window)
                    widget = indexeditem(window.widgets, focus_widget) 
    
                    //set value
                    widget.value = (MouseY - drop_box_y) / drop_height
                    listopt = indexeditem(widget.listopts, widget.value)
                    widget.txt = listopt.txt

                    //dropdown select event
                    window_ref = focus_window
                    widget_ref = focus_widget
                    result = TRUE
        
                    //redraw widget
                    DrawWidget(focus_window, focus_widget)   
        
                    //clear drop box
                    FreeSprite(drop_box)
                    drop_box = NULL
                    focus_window = -1
                    focus_widget = -1 
                endif
                //stop here to avoid interaction with widgets underneath the drop box
                exit
            else
                //clear drop box
                FreeSprite(drop_box)
                drop_box = NULL
                focus_window = -1
                focus_widget = -1    
            endif
            
        else    
        
            //textbox   
            if EventKey then
                //keyboard input event
                window_ref = focus_window
                widget_ref = focus_widget
                result = TRUE
                //stop here!
                exit
            endif 
            
        endif
    endif    
    
    //*** mouse handler ***
    
    //find which window is active
    loop window through windows
        if window.visible then
            if PointInRect(MouseX, MouseY, window.x, window.y, window.width, window.height) then 

                // *** window selected ***
                window_ref = window.ref 
                
                //find which widget is active 
                loop widget through window.widgets
                    if widget.visible and widget.enabled then
                        if PointInRect(MouseX, MouseY, window.x + widget.x, window.y + widget.y, widget.width, widget.height) then
                        
                            // *** widget selected *** 
                            widget_ref = widget.ref      
                            
                            //mouse events
                            case mouse_state of
                            (MOUSE_UP):
                                //set hover if not already
                                if (mouse_window <> window.ref) or (mouse_widget <> widget.ref) then
                                    //clear old hover if set to another widget
                                    if mouse_window > -1 then DrawWidget(mouse_window, mouse_widget)
                                    //set new hover                              
                                    mouse_window = window.ref
                                    mouse_widget = widget.ref
                                    DrawWidget(mouse_window, mouse_widget, TRUE)
                                endif
                            
                            (MOUSE_CLICK):
                                //clear hover if changed
                                if mouse_window > -1 then 
                                    if (mouse_window <> window.ref) or (mouse_widget <> widget.ref) then DrawWidget(mouse_window, mouse_widget)
                                endif                            
                            
                                //hold
                                mouse_window = window.ref
                                mouse_widget = widget.ref
                                DrawWidget(mouse_window, mouse_widget, FALSE, TRUE)       
                                
                                //clear focus if changed
                                if focus_window > -1 then
                                    if (focus_window <> window.ref) or (focus_widget <> widget.ref) then 
                                        temp_window = focus_window
                                        temp_widget = focus_widget                              
                                        focus_window = -1
                                        focus_widget = -1   
                                        DrawWidget(temp_window, temp_widget)
                                    endif
                                endif
                                 
                            (MOUSE_RELEASE):
                                //check if mouse has moved to a different widget 
                                if mouse_window > -1 then
                                    if (mouse_window <> window.ref) or (mouse_widget <> widget.ref) then
                                        //clear hold visual and target
                                        DrawWidget(mouse_window, mouse_widget)                                                                            
                                    else
                                    
                                        // *** process mouse release event *** 
                                        
                                        EventMouse(window, widget) 
                                        //return 'click on something' event                                        
                                        result = TRUE
                                    endif                                    
                                endif   
                                
                                //clear hold / set hover
                                mouse_window = window.ref
                                mouse_widget = widget.ref
                                //draw widget
                                DrawWidget(window.ref, widget.ref, TRUE, FALSE)                                
                                
                            (MOUSE_DOWN):  
                                //check if mouse has moved to a different widget
                                if (mouse_window <> window.ref) or (mouse_widget <> widget.ref) then
                                    //clear hold visual
                                    if mouse_window > -1 then
                                        DrawWidget(mouse_window, mouse_widget)
                                        mouse_window = -1
                                        mouse_widget = -1
                                    endif  
                                else    
                                
                                    //sliders can be dragged with the mouse down
                                    if widget.mode = WIDGET_SLIDE then
                                        //process event
                                        EventMouse(window, widget)   
                                        //draw updated position
                                        DrawWidget(mouse_window, mouse_widget, FALSE, TRUE)   
                                        //return 'click on something' event                                        
                                        result = TRUE
                                    endif
                                endif 
                            endcase 
                            
                            // EXIT function
                            exit
                            
                        endif
                    endif
                endloop
            
            endif
        endif
    endloop
    
    // *** no widget selected ***    
    
    //clear hover and hold (visual) if set                  
    if mouse_window > -1 then
        DrawWidget(mouse_window, mouse_widget)
        mouse_window = -1
        mouse_widget = -1
    endif 
    
    //clear focus on click
    case mouse_state of 
    (MOUSE_CLICK):
        if focus_window > -1 then
            temp_window = focus_window
            temp_widget = focus_widget                              
            focus_window = -1
            focus_widget = -1   
            DrawWidget(temp_window, temp_widget)
        endif
    endcase
    
end

procedure EventMouse(window:^gui_window, widget:^gui_widget)

//  * Called by OpalEvent
//  * Handles mouse-driven widget events

var
    other_widget    : ^gui_widget
    listopt         : ^gui_listopt 

begin   
    
    case widget.mode of
    (WIDGET_CHECK):
        //toggle check
        widget.value = -widget.value - 1 
                                    
    (WIDGET_RADIO),(WIDGET_TOGGLE):
        if widget.group > -1 then
            //if grouped, unset other radios in group
            loop other_widget through window.widgets                                             
                if other_widget.group = widget.group then
                    other_widget.value = FALSE    
                    //draw other widget
                    DrawWidget(window.ref, other_widget.ref)
                endif                 
            endloop
            //set radio      
            widget.value = TRUE
        else
            //if ungrouped, simply toggle
            widget.value = -widget.value - 1                                         
        endif       
                                       
    (WIDGET_TEXT):
        //focus on Text boxes
        focus_window = window.ref
        focus_widget = widget.ref      
        
    (WIDGET_SLIDE):
        //set value to mouse position
        widget.value = ((toreal(MouseX) - widget.x - window.x - (slide_width / 2)) * (widget.value_max - widget.value_min) / (widget.width - slide_width)) + widget.value_min
        //bound Min / max
        widget.value = Max(widget.value, widget.value_min)
        widget.value = Min(widget.value, widget.value_max) 
        
    (WIDGET_DROP):  
        CreateDropbox(window, widget)    
        
    (WIDGET_LIST): 
        //set value
        widget.value = (MouseY - window.y - widget.y) / text_height
        //bound 
        widget.value = Min(widget.value, widget.value_max - 1)
        //set Text
        listopt = indexeditem(widget.listopts, widget.value)
        widget.txt = listopt.txt
    endcase
end 

function EventKey : boolean

//  * handle keyboard input into focussed Text widget

var
    window  : ^gui_window 
    widget  : ^gui_widget
    key     : integer 
    len     : integer
     
begin
    //get key    
    key = GetKey
    if key > 0 then
    
        //set pointers
        window = indexeditem(windows, focus_window)
        widget = indexeditem(window.widgets, focus_widget) 

        len = Length(widget.txt)
        
        conditions
        (key = $08): // backspace
            if len > 0 then widget.txt = Left(widget.txt, len - 1)
        (key >= $20): // printable characters
            if len < widget.value_max then widget.txt = widget.txt + Chr(key)
        endconditions 
        
        //set 'value' of Text - may cause Warnings with debugger on (harmless)
        widget.value = toint(widget.txt)
        
        //redraw
        DrawWidget(window.ref, widget.ref)  
        
        //return keyboard event
        result = TRUE
    endif        
end

function CreateWidget(window_ref:integer, mode:integer, x:integer, y:integer, width:integer, height:integer, icon_image:element, value:integer, value_min:integer, value_max:integer, txt:string, group:integer, enabled:boolean, visible:boolean, textcolor:integer) : integer

//  * Create a widget in the window referenced. Returns handle of widget.
//  * Can create any widget type; called by other Create functions.
//  * NOT Exported - bad values may break Opalescence!

var
    window      : ^gui_window 
    widget      : ^gui_widget
    listopt     : ^gui_listopt
    
begin
    //set pointers
    window = indexeditem(windows, window_ref)
            
    //create widget
    widget           = newitem(window.widgets)
    widget.mode      = mode
    widget.ref       = window.widget_count
    widget.x         = x
    widget.y         = y
    widget.width     = width
    widget.height    = height
    if icon_image <> NULL then widget.icon      = icon_image  
    widget.group     = group
    widget.value     = value 
    widget.value_min = value_min
    widget.value_max = value_max   
    widget.txt       = txt
    widget.enabled   = enabled
    widget.visible   = visible
    widget.textcolor = textcolor

    //create first option if listbox / dropdown    
    case mode of
    
    (WIDGET_DROP),(WIDGET_LIST):   
        listopt = newitem(widget.listopts)
        listopt.txt = txt
        listopt.ref = 0            
        //count
        widget.value_max = 1
    
    endcase

    //render widget            
    DrawWidget(window.ref, widget.ref)
            
    //return ref
    result = window.widget_count       
    inc(window.widget_count)
end

procedure CreateDropbox(window:^gui_window, widget:^gui_widget)
    
begin
    //focus on parent
    focus_window = window.ref
    focus_widget = widget.ref

    //set position
    drop_box_x      = window.x + widget.x  
    drop_box_y      = window.y + widget.y //+ widget.height
    drop_box_width  = widget.width
    drop_box_height = drop_height * widget.value_max

    //create the dropbox window
    drop_box = CreateSprite(drop_box_width, drop_box_height)        
    PositionSprite(drop_box, drop_box_x, drop_box_y)
    SpriteIndex(drop_box, window_zorder + window_count)
    SetFont(opal_fontface, opal_fontsize, opal_fontstyle, drop_box)       
        
    //draw dropbox
    DrawDropbox 
end

procedure DrawWindow(window_ref:integer)

//  * Render window from scratch

var
    window : ^gui_window
    widget : ^gui_widget
begin
    //set pointers
    window = indexeditem(windows, window_ref)

    //background and border
    case window.backcolor of
        //not set, use theme color
        DEFAULT_COLOR: CLS(opal_window, window.sprite)
        //use custom color
        default: CLS(window.backcolor, window.sprite)
    endcase    
    Rect(0, 0, window.width, window.height, opal_windowborder, FALSE, window.sprite)
    
    //texture
    if window.texture <> NULL then TileImage(window.texture, 0, 0, window.width, window.height, window.sprite)
            
    //render widgets
    loop widget through window.widgets
        DrawWidget(window.ref, widget.ref) 
    endloop
end

Procedure DrawWidget(window_ref:integer, widget_ref:integer, hover:boolean = FALSE, hold:boolean = FALSE)

//  * Render widget on window                                                                  

var
    window  : ^gui_window
    widget  : ^gui_widget
    listopt : ^gui_listopt

    x, y    : integer    
    state   : integer
    
begin
 
    //set pointers
    window = indexeditem(windows, window_ref)
    widget = indexeditem(window.widgets, widget_ref)
    
    //set viewport
    ClipRect(widget.x, widget.y, widget.x + widget.width, widget.y + widget.height, window.sprite)
    
    //blit out widget
    case window.backcolor of
        //not set, use theme color
        DEFAULT_COLOR: CLS(opal_window, window.sprite)
        //use custom color
        default: CLS(window.backcolor, window.sprite)
    endcase
    //apply texture
    if window.texture <> NULL then TileImage(window.texture, 0, 0, window.width, window.height, window.sprite)

    if widget.visible    
        /// *** prep stage *** 
        
        //select widget state for graphics 
        case widget.mode of
        
        (WIDGET_CHECK),(WIDGET_RADIO): 
            if widget.value <> 0      then state = state + WIDGET_SET
            if widget.enabled = FALSE then state = state + WIDGET_DISABLED
            if hover = TRUE           then state = state + WIDGET_HOVER
            if hold = TRUE            then state = state + WIDGET_HOLD
        
        (WIDGET_SLIDE),(WIDGET_DROP):
            if widget.enabled = FALSE then state = state + WIDGET_DISABLED/2
            if hover = TRUE           then state = state + WIDGET_HOVER/2
            if hold = TRUE            then state = state + WIDGET_HOLD/2
            
        endcase
        
        //set pen colour
        if widget.textcolor = DEFAULT_COLOR then
            //use theme colour
            case widget.mode of 
        
            (WIDGET_LABEL),(WIDGET_CHECK),(WIDGET_RADIO):
                // label text
                if widget.enabled then                   
                    PenColor(opal_label, window.sprite)
                else
                    PenColor(opal_labeldisabled, window.sprite)
                endif  

            (WIDGET_BUTTON),(WIDGET_TOGGLE): 
                // button text
                if widget.enabled then                   
                    PenColor(opal_buttontext, window.sprite)
                else
                    PenColor(opal_buttontextdis, window.sprite)
                endif  

            (WIDGET_TEXT),(WIDGET_DROP),(WIDGET_LIST): 
                // widget text
                if widget.enabled then                   
                    PenColor(opal_widgettext, window.sprite)
                else
                    PenColor(opal_widgettextdis, window.sprite)
                endif                           
            endcase 
             
        else
            //custom Text colour
            PenColor(widget.textcolor, window.sprite)
        endif      

        // *** render stage ***
                
        //draw widget according to mode
        case widget.mode of

        (WIDGET_LABEL):
            //icon
            if widget.icon<>NULL then DrawImage(widget.icon, widget.x, widget.y, window.sprite)
            //label
            Text(widget.x, widget.y, widget.txt, window.sprite)

        (WIDGET_CHECK):
            // draw checkbox       
            DrawImage(gfx_check[state], widget.x, widget.y, window.sprite)
            //icon
            if widget.icon<>NULL then DrawImage(widget.icon, widget.x + label_space, widget.y, window.sprite)
            //label
            Text(widget.x + label_space, widget.y, widget.txt, window.sprite)
            
        (WIDGET_RADIO):
            // draw radio button
            DrawImage(gfx_radio[state], widget.x, widget.y, window.sprite)
            //icon
            if widget.icon<>NULL then DrawImage(widget.icon, widget.x + label_space, widget.y, window.sprite)
            // label       
            Text(widget.x + label_space, widget.y, widget.txt, window.sprite)

        (WIDGET_BUTTON),(WIDGET_TOGGLE):
            // draw button       
            if widget.enabled then 
                //enabled
                if hold or widget.value= -1 then       
                    //held down, or set (if a pushbutton)
                    Rect(widget.x, widget.y, widget.width, widget.height, opal_buttonhold, TRUE, window.sprite)
                else  
                    //not set
                    Rect(widget.x, widget.y, widget.width, widget.height, opal_button, TRUE, window.sprite)
                endif
                Rect(widget.x, widget.y, widget.width, widget.height, opal_windowborder, FALSE, window.sprite)
            else   
                //disabled
                Rect(widget.x, widget.y, widget.width, widget.height, opal_buttondisabled, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_buttonbordis, FALSE, window.sprite)
            endif   
            // hover
            if hover then Rect(widget.x+1, widget.y+1, widget.width-2, widget.height-2, opal_buttonhover, FALSE, window.sprite)
            // position and draw icon, if present
            if widget.icon<>NULL then
                x = widget.x + (widget.width - ImageWidth(widget.icon)) / 2
                y = widget.y + (widget.height - Imageheight(widget.icon)) / 2
                DrawImage(widget.icon, x, y, window.sprite) 
            endif
            // position and draw label   
            x = widget.x + (widget.width - TextWidth(widget.txt, window.sprite)) / 2
            y = widget.y + (widget.height - text_height) / 2 
            Text(x, y, widget.txt, window.sprite)

        (WIDGET_TEXT):
            // draw box      
            if widget.enabled then 
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widget, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetborder, FALSE, window.sprite)
            else
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetdisabled, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetbordis, FALSE, window.sprite)                
            endif   
            // hover / hold
            if hover then Rect(widget.x+1, widget.y+1, widget.width-2, widget.height-2, opal_widgethover, FALSE, window.sprite)

            // draw text   
            x = widget.x + 1
            y = widget.y + (widget.height - text_height) / 2  
            
            if (focus_window = window.ref) and (focus_widget = widget.ref) then              
                //draw Text with caret if focussed
                //if Text doesn't fit in the box, align Right
                Text(x - Max(TextWidth(widget.txt+"|", window.sprite) - widget.width, 0), y, widget.txt+"|", window.sprite)
            else   
                //Textbox contents
                Text(x, y, widget.txt, window.sprite)
            endif  
            
        (WIDGET_SLIDE): 
            //slider box
            if widget.enabled then                
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widget, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetborder, FALSE, window.sprite)
            else
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetdisabled, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetbordis, FALSE, window.sprite)
            endif   

            //calc position of slider
            x = ((toreal(widget.value) - widget.value_min) * (widget.width - slide_width) / (widget.value_max - widget.value_min)) + widget.x
            //draw slider
            DrawImage(gfx_slide[state], x, widget.y, window.sprite)        
            
        (WIDGET_DROP):
            // draw box      
            if widget.enabled then 
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widget, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetborder, FALSE, window.sprite)
            else
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetdisabled, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetbordis, FALSE, window.sprite)                
            endif   

            // draw text   
            x = widget.x + 1
            y = widget.y + (widget.height - text_height) / 2  
            Text(x, y, widget.txt, window.sprite)
            
            //graphic
            DrawImage(gfx_drop[state], widget.x + widget.width - drop_width, widget.y, window.sprite) 

        (WIDGET_LIST):
            // draw box      
            if widget.enabled then 
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widget, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetborder, FALSE, window.sprite)
            else
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetdisabled, TRUE, window.sprite)
                Rect(widget.x, widget.y, widget.width, widget.height, opal_widgetbordis, FALSE, window.sprite)                
            endif
               
            // hover
            if hover then Rect(widget.x+1, widget.y+1, widget.width-2, widget.height-2, opal_widgethover, FALSE, window.sprite)

            // draw text   
            x = widget.x + 1
            y = 0
            
            loop listopt through widget.listopts        
                //hilight selected line
                if widget.value = listopt.ref then Rect(x, widget.y + (y * text_height) + 1, widget.width - 2, text_height - 2, opal_widgetselect, TRUE, window.sprite)
                //text
                Text(x, widget.y + (y * text_height), listopt.txt, window.sprite)
                //next line                
                inc(y)
            endloop 

        endcase
    endif 
    
    //clear view port
    ClipClear(window.sprite)
    
    //update window
    //SpriteRefresh(window.sprite) 
end

procedure DrawDropbox
var
    y       : integer
    text_y  : integer 
    window  : ^gui_window    
    widget  : ^gui_widget
    listopt : ^gui_listopt 
begin
    //set pointers
    window = indexeditem(windows, focus_window)
    widget = indexeditem(window.widgets, focus_widget) 
    
    //draw box
    CLS(opal_widget,drop_box)
    Rect(0, 0, drop_box_width, drop_box_height, opal_widgetborder, FALSE, drop_box)
    Rect(1, 1, drop_box_width - 2, drop_box_height - 2, opal_widgethover, FALSE, drop_box)
        
    //draw options 
    PenColor(opal_widgettext, drop_box)
    text_y = (drop_height - text_height) / 2
    loop listopt through widget.listopts        
        //hilight selected line
        if widget.value = listopt.ref then Rect(1, y * drop_height + 1, drop_box_width - 2, drop_height - 2, opal_widgetselect, TRUE, drop_box)
        //text
        Text(1, y * drop_height + text_y, listopt.txt, drop_box)
        //next line                
        inc(y)
    endloop 
end

begin
end
